//*****************************************************//
//Dateiname: CreateTree.c
//Autor: Ptschner Stephan
//Projekt: niCE - nice insignificant compression engine
//Zweck: Funktionen zum Erzeugen eines dynamischen Baumes und einer Liste
//Datum der letzten Aenderung: 09.05.2003
//*****************************************************//


//Eigene spezifische INCLUDES
#include "../nice.h"
#include "../huffman.h"

//***************************************************************************
//
//					Funktionen
//
//***************************************************************************

/**
 *
 *	@function: Hauptfunktion, die von Datei CreateTree.c zur Verfgung gestellt wird
 *				erzeugt automatisch einen Huffman-Baum aus den Informationen aus dem Feld
 *	@params:
 *				pt_array_Probabilities ... Zeiger auf das erste Element eines 256 Felder grossen double-Feldes
 *	@return value:
 *				RootKnoten des erzeugten Baumes
 *
 */
Node *CreateTree(double *pt_array_Probabilities)
{

	Node *pt_BeginOfList = CreateList (pt_array_Probabilities); // Erzeugt eine dynamische Liste aus dem Array


	if(pt_BeginOfList == NULL || pt_BeginOfList->pt_Next == NULL)
		return pt_BeginOfList ;



	while (pt_BeginOfList != NULL) // Schleife solange durchfhren bis ganze Liste bearbeitet wurde
	{
		Node *pt_RightNode;		// neuer einzufgender rechter Knoten
		Node *pt_LeftNode;		// neuer einzufgender linker Knoten
		Node *pt_NewNode;		// neuer "verschmolzener" Knoten
		
		// die beiden niederwertigsten Knoten werden aus der Liste entfernt und zu einem neuen Knoten verschmolzen
		pt_RightNode = pt_BeginOfList;		// niederwertigster Knoten ist rechtes Kind des neuen Knotens
		pt_BeginOfList = pt_BeginOfList->pt_Next;	// Knoten wird aus der Liste gelscht
		pt_RightNode->pt_Next = NULL;				// Verknpfung des rechten Kindes wird gelscht


		// gleiches Verfahren wie bei rechtem Kind
		pt_LeftNode = pt_BeginOfList;
		pt_BeginOfList = pt_BeginOfList->pt_Next;
		pt_LeftNode->pt_Next = NULL;

		// Neuer Knoten wird erzeugt: Verknpfung zu linkem und rechtem Kind
		pt_NewNode = CreateNode( ROOTNODE,		// iSimpleNode wird auf "0" gesetzt, da es sich nicht um einen einfachen Knoten sondern um einen verschmolzenen handelt
								'\0',	// Pseudoeingabe, da ein verschmolzener Knoten keinen Buchstabeninhalt hat
								pt_RightNode->ldProbability + pt_LeftNode->ldProbability,	// Wahrscheinlichkeit des neuen Knotens setzt sich aus den Wahrscheinlichkeiten des Kindknotens zusammen
								NULL,			// Vaterknoten kann noch nicht angegeben werden
								pt_LeftNode,	// Verknpfung zum linken Kindknoten
								pt_RightNode,	// Verknpfung zum rechten Kindknoten
								NULL);			// Da sich ein "verschmolzener Knoten in keiner Liste mehr befinden soll, wird kein Link erstellt
		
		//Verweise der Kinderknoten auf den gerade erzeugten Vater erstellen
		pt_RightNode->pt_Up = pt_NewNode;
		pt_LeftNode->pt_Up = pt_NewNode;

		
		if (pt_BeginOfList != NULL) // Solange BeginOfList nicht ein NULL-Knoten ist, ...
			pt_BeginOfList = InsertNode (pt_BeginOfList, pt_NewNode); // ... sollen die neu erzeugten Knoten in die Liste eingefgt werden, falls dies nicht der falls ist, ...
		else
			return pt_NewNode;	// ... ist das Programm bereits am Wurzelknoten angekommen, dh. der Wurzelknoten kann zurckgegeben werden


	}

	return NULL; // Pseudorckgabewert, falls Algorithmus falsch beendet wurde
}

/**
 *
 *	@function: Hauptfunktion, die von Datei CreateTree.c zur Verfgung gestellt wird
 *				erzeugt automatisch einen Huffman-Baum aus den Informationen aus einer Datei
 *	@params:
 *				pt_File ... Zeiger auf die geffnete Datei
 *				pt_BufferSequence ... Buffer der aus der Datei bereits ausgelesen wurde, jedoch nicht verwertet werden konnte
 *				
 *	@return value:
 *				Wurzelknoten des erzeugten Baumes
 */
Node *CreateTreeFromFile(FILE *pt_File, BitSequence *pt_BufferSequence)
{
	Node *pt_RootNode = NULL;
	char temp_SimpleNode = 0;
	temp_SimpleNode = GetBitFromFile(pt_File, pt_BufferSequence);
	if (temp_SimpleNode != -2)
	{
		int temp_cContent = (temp_SimpleNode == LEAFNODE ?  GetByteFromFile(pt_File, pt_BufferSequence) : 0);			// Falls es sich um einen Blattknoten handelt, so entscheidet das nchste Byte den Inhalt, ansonsten wird '\0' gesetzt
		// Erster Knoten wird erzeugt
		pt_RootNode = CreateNode((int) temp_SimpleNode,// iSimpleNode wird auf den eingelesenen Wert aus der Datei gesetzt
								(char) temp_cContent,
								0.,				// Wahrscheinlichkeit auf 0, da fr Dekomprimierung nicht von Interesse
								NULL,			// Vaterknoten kann noch nicht angegeben werden
								NULL,			// Verknpfung zum linken Kindknoten (Beim Auslesen des ersten Knotens noch unbekannt)
								NULL,			// Verknpfung zum rechten Kindknoten (Beim Auslesen des ersten Knotens noch unbekannt)
								NULL);			// Da sich ein ausgelesenerKnoten in keiner Liste befinden soll, wird kein Link erstellt
		if (temp_SimpleNode == 0)
			TraverseTreeforRead(pt_RootNode, "", &ReadLeftChildFromFile, &ReadRightChildFromFile, pt_BufferSequence, pt_File);

	}
	return pt_RootNode;

}


/**
 *
 *	@function: Lscht die Knoten und gibt den reservierten Speicher wieder frei
 *	@params:
 *				pt_rootNode ... Zeiger auf den Wurzelknoten ab welchem die BAumstruktur gelscht werden soll
 *	@return value:
 *				KEINER
 *
 */
void DestroyTree(Node *pt_rootNode)
{
	TraverseTree(pt_rootNode, "", &DeleteNode, NULL);
	pt_rootNode = NULL;
}


/**
 *
 *	@function: Erzeugt eine Liste mit den Inhalten aus dem Array, jedoch geordnet nach Wahrscheinlichkeiten und ASCII-Code
 *	@params:
 *				pt_array_Probabilities ... Zeiger auf das erste Element eines 256 Felder grossen double-Feldes
 *	@return value:
 *				Beginn der Liste (Knoten mit geringster Wahrscheinlichkeit am Beginn)
 *
 */
Node *CreateList (double *pt_array_Probabilities)
{
	Node *pt_BeginOfList = NULL;	// Listenbeginn wird mit einem Nullpointer initialisiert
	int run_a;						// Laufvariable

	// Durchlaufen des bergebenen Feldes und erstellen einer dynamischen Liste mit gleichem Inhalt
	for (run_a = 255; run_a >= 0; run_a--)
	{
		if (pt_array_Probabilities[run_a] != 0.)	// Feld soll nur als Knoten aufgenommen werden, wenn dessen Wahrscheinlichkeit grsser 0 ist
		{
			Node *pt_NewNode = CreateNode(LEAFNODE, (char)run_a, pt_array_Probabilities[run_a], NULL, NULL, NULL, NULL);	// Erzeugen eines neuen Knotens mit dem Inhalt aus dem aktuellen Feldelement
			pt_BeginOfList = InsertNode (pt_BeginOfList, pt_NewNode);	// Einfgen neuen Knotens an der richtigen Position
		}
	}
	return pt_BeginOfList;
}

/**
 *
 *	@function: Erzeugt einen dynamischen Knoten der Struktur "Node" und belegt diesen mit Werten aus den Parametern
 *	@params:
 *				cContent ... Buchstabe, welcher im Knoten gespeichert werden soll
 *				ldProbability ... Wahrscheinlichkeit mit der der Buchstabe cContent im Text auftritt
 *				pt_Up, pt_DownLeft, pt_DownRight ... Links in der Baumstruktur
 *				pt_Next ... Link in der Listenstruktur
 *	@return value:
 *				Neu erzeugter Knoten, mit Infos aus Parametern als Inhalt
 *
 */
Node *CreateNode (int iSimpleNode, char cContent, double ldProbability, Node *pt_Up, Node *pt_DownLeft, Node *pt_DownRight, Node *pt_Next)
{
	Node *pt_NewNode = (Node *) malloc (sizeof(Node));		// Erzeugen eines neuen Knotens

	// Zuweisen der Informationen aus den Parametern an den Knoten
		// Baumstruktur
	pt_NewNode->pt_Up = pt_Up;	// Link zum Vaterknoten
	pt_NewNode->pt_Down[LEFT] = pt_DownLeft;	// Link zum Kindknoten links
	pt_NewNode->pt_Down[RIGHT] = pt_DownRight;	//Link zum Kindknoten rechts
		//Listenstruktur
	pt_NewNode->pt_Next = pt_Next;				//Links zum Nachbarn
		//Inhalt
	pt_NewNode->iSimpleNode = iSimpleNode;		// Kennzeichnung, ob es sich um einen einfachen Knoten handelt oder einen bereits aus zwei verschmolzenem
	pt_NewNode->cContent = cContent;			// Buchstabe
	pt_NewNode->ldProbability = ldProbability;	// Auftrittswahrscheinlichkeit des Buchstaben
	
	
	return pt_NewNode;
}


/**
 *
 *	@function: Fgt einen bereits bestehenden Knoten n der richtigen Position in der Liste ein
 *	@params:
 *				pt_BeginOfList ... Zeiger auf den Beginn der Liste
 *				pt_InsertNew ... Zeiger auf den einzufgenden Knoten (ACHTUNG: next Element wird berschrieben!!!)
 *	@return value:
 *				Beginn der Liste mit dem hinzugefgten Element
 *
 */
Node *InsertNode (Node *pt_BeginOfList, Node *pt_InsertNew)
{
	Node *pt_ActualNode = pt_BeginOfList;	// Zeiger, welcher auf der Suche nach der richtigen Position die Liste durchluft

	//Sonderbehandlung: pt_BeginOfList zeigt noch auf "NULL" -> neues Element ist erstes Element der Liste
	if (pt_BeginOfList == NULL)
		return pt_InsertNew;

	// Sonderbehandlung: Position des neuen Elements soll an erster Stelle sein
	if ((pt_BeginOfList->ldProbability > pt_InsertNew->ldProbability)		// Wahrscheinlichkeit des neuen Elements ist geringer als des derztg. ersten
			|| (pt_BeginOfList->ldProbability == pt_InsertNew->ldProbability && // Wahrscheinlichkeit des neuen Elements ist gleich der Wahrscheinlichkeit des ersten Elements und ...
			(pt_ActualNode->iSimpleNode == ROOTNODE || pt_BeginOfList->cContent < pt_InsertNew->cContent))) // ... (neues Element ist kein "SimpleNode" oder Wertigkeit des Buchstaben ist geringer)
	{
		pt_InsertNew->pt_Next = pt_BeginOfList;
		return pt_InsertNew;
	}

	// Einfgen des Knotens falls ...
	while (1)
	{
		if (   (pt_ActualNode->pt_Next == NULL)		// bereits der vorletzte Knoten erreicht wurde oder ...
			|| (pt_ActualNode->pt_Next->ldProbability > pt_InsertNew->ldProbability) // die Wahrscheinlichkeit des neuen Elements geringer ist als des derztg. ersten oder
			|| (pt_ActualNode->pt_Next->ldProbability == pt_InsertNew->ldProbability // die Wahrscheinlichkeit des neuen Elements gleich der Wahrscheinlichkeit des ersten Elements ist und ...
					&& (pt_ActualNode->iSimpleNode == ROOTNODE || pt_ActualNode->pt_Next->cContent < pt_InsertNew->cContent ))) // ... (das neue Element ist kein "SimpleNode" oder die Wertigkeit des Buchstaben ist geringer als die des ersten)
		{
			pt_InsertNew->pt_Next = pt_ActualNode->pt_Next;		// Insert-Knoten wird mit dem hinteren Teil verknpft (setzen des next)
			pt_ActualNode->pt_Next = pt_InsertNew;				// Insert-Knoten wird endgltig eingefgt hinter der aktuellen Stelle
			return pt_BeginOfList;
		}
		pt_ActualNode = pt_ActualNode->pt_Next;				// nchster Schritt in der Liste beim Suchen der richtigen Position
	}

	return NULL;	// Fehlerwertrckgabe, da alle Mglichkeiten abgedeckt sein mssten

}